#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from io import StringIO
from unittest import mock

import pytest

from airflow.providers.ftp.hooks import ftp as fh


class TestFTPHook:
    def setup_method(self):
        self.path = "/some/path"
        self.conn_mock = mock.MagicMock(name="conn")
        self.get_conn_orig = fh.FTPHook.get_conn

        def _get_conn_mock(hook):
            hook.conn = self.conn_mock
            return self.conn_mock

        fh.FTPHook.get_conn = _get_conn_mock

    def teardown_method(self):
        fh.FTPHook.get_conn = self.get_conn_orig

    def test_close_conn(self):
        ftp_hook = fh.FTPHook()
        ftp_hook.get_conn()
        ftp_hook.close_conn()

        self.conn_mock.quit.assert_called_once_with()

    def test_describe_directory(self):
        with fh.FTPHook() as ftp_hook:
            ftp_hook.describe_directory(self.path)

        self.conn_mock.mlsd.assert_called_once_with(self.path)

    def test_list_directory(self):
        with fh.FTPHook() as ftp_hook:
            ftp_hook.list_directory(self.path)

        self.conn_mock.nlst.assert_called_once_with(self.path)

    def test_create_directory(self):
        with fh.FTPHook() as ftp_hook:
            ftp_hook.create_directory(self.path)

        self.conn_mock.mkd.assert_called_once_with(self.path)

    def test_delete_directory(self):
        with fh.FTPHook() as ftp_hook:
            ftp_hook.delete_directory(self.path)

        self.conn_mock.rmd.assert_called_once_with(self.path)

    def test_delete_file(self):
        with fh.FTPHook() as ftp_hook:
            ftp_hook.delete_file(self.path)

        self.conn_mock.delete.assert_called_once_with(self.path)

    def test_rename(self):
        from_path = "/path/from"
        to_path = "/path/to"
        with fh.FTPHook() as ftp_hook:
            ftp_hook.rename(from_path, to_path)

        self.conn_mock.rename.assert_called_once_with(from_path, to_path)
        self.conn_mock.quit.assert_called_once_with()

    def test_mod_time(self):
        self.conn_mock.sendcmd.return_value = "213 20170428010138"

        path = "/path/file"
        with fh.FTPHook() as ftp_hook:
            ftp_hook.get_mod_time(path)

        self.conn_mock.sendcmd.assert_called_once_with("MDTM " + path)

    def test_mod_time_micro(self):
        self.conn_mock.sendcmd.return_value = "213 20170428010138.003"

        path = "/path/file"
        with fh.FTPHook() as ftp_hook:
            ftp_hook.get_mod_time(path)

        self.conn_mock.sendcmd.assert_called_once_with("MDTM " + path)

    def test_get_size(self):
        self.conn_mock.size.return_value = 1942

        path = "/path/file"
        with fh.FTPHook() as ftp_hook:
            ftp_hook.get_size(path)

        self.conn_mock.size.assert_called_once_with(path)

    def test_retrieve_file(self):
        _buffer = StringIO("buffer")
        with fh.FTPHook() as ftp_hook:
            ftp_hook.retrieve_file(self.path, _buffer)
        self.conn_mock.retrbinary.assert_called_once_with("RETR /some/path", _buffer.write, 8192)

    def test_retrieve_file_with_callback(self):
        func = mock.Mock()
        _buffer = StringIO("buffer")
        with fh.FTPHook() as ftp_hook:
            ftp_hook.retrieve_file(self.path, _buffer, callback=func)
        self.conn_mock.retrbinary.assert_called_once_with("RETR /some/path", func, 8192)

    def test_connection_success(self):
        with fh.FTPHook() as ftp_hook:
            status, msg = ftp_hook.test_connection()
            assert status is True
            assert msg == "Connection successfully tested"

    def test_connection_failure(self):
        self.conn_mock = mock.MagicMock(name="conn_mock", side_effect=Exception("Test"))
        fh.FTPHook.get_conn = self.conn_mock
        with fh.FTPHook() as ftp_hook:
            status, msg = ftp_hook.test_connection()
            assert status is False
            assert msg == "Test"


class TestIntegrationFTPHook:
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        from airflow.models import Connection

        create_connection_without_db(
            Connection(conn_id="ftp_passive", conn_type="ftp", host="localhost", extra='{"passive": true}')
        )

        create_connection_without_db(
            Connection(conn_id="ftp_active", conn_type="ftp", host="localhost", extra='{"passive": false}')
        )

        create_connection_without_db(
            Connection(
                conn_id="ftp_custom_port",
                conn_type="ftp",
                host="localhost",
                port=10000,
                extra='{"passive": true}',
            )
        )

        create_connection_without_db(
            Connection(
                conn_id="ftp_custom_port_and_login",
                conn_type="ftp",
                host="localhost",
                port=10000,
                login="user",
                password="pass123",
                extra='{"passive": true}',
            )
        )

        create_connection_without_db(
            Connection(
                conn_id="ftp_encoding",
                conn_type="ftp",
                host="localhost",
                extra='{"encoding": "cp1251"}',
            )
        )

    def _test_mode(self, hook_type, connection_id, expected_mode):
        hook = hook_type(connection_id)
        conn = hook.get_conn()
        conn.set_pasv.assert_called_once_with(expected_mode)

    @mock.patch("ftplib.FTP")
    def test_ftp_passive_mode(self, mock_ftp):
        from airflow.providers.ftp.hooks.ftp import FTPHook

        self._test_mode(FTPHook, "ftp_passive", True)

    @mock.patch("ftplib.FTP")
    def test_ftp_active_mode(self, mock_ftp):
        from airflow.providers.ftp.hooks.ftp import FTPHook

        self._test_mode(FTPHook, "ftp_active", False)

    @mock.patch("ftplib.FTP")
    def test_ftp_custom_port(self, mock_ftp):
        from airflow.providers.ftp.hooks.ftp import FTPHook

        hook = FTPHook("ftp_custom_port")
        conn = hook.get_conn()
        conn.connect.assert_called_once_with("localhost", 10000)
        conn.login.assert_not_called()
        conn.set_pasv.assert_called_once_with(True)

    @mock.patch("ftplib.FTP")
    def test_ftp_custom_port_and_login(self, mock_ftp):
        from airflow.providers.ftp.hooks.ftp import FTPHook

        hook = FTPHook("ftp_custom_port_and_login")
        conn = hook.get_conn()
        conn.connect.assert_called_once_with("localhost", 10000)
        conn.login.assert_called_once_with("user", "pass123")
        conn.set_pasv.assert_called_once_with(True)

    @mock.patch("ftplib.FTP_TLS")
    def test_ftps_passive_mode(self, mock_ftp):
        from airflow.providers.ftp.hooks.ftp import FTPSHook

        self._test_mode(FTPSHook, "ftp_passive", True)

    @mock.patch("ftplib.FTP_TLS")
    def test_ftps_active_mode(self, mock_ftp):
        from airflow.providers.ftp.hooks.ftp import FTPSHook

        self._test_mode(FTPSHook, "ftp_active", False)

    @mock.patch("ftplib.FTP")
    def test_ftp_encoding_extra(self, mock_ftp):
        from airflow.providers.ftp.hooks.ftp import FTPHook

        hook = FTPHook("ftp_encoding")
        hook.get_conn()
        assert mock_ftp.mock_calls[0] == mock.call(encoding="cp1251")

    @mock.patch("ftplib.FTP_TLS")
    def test_ftps_encoding_extra(self, mock_ftp_tls):
        from airflow.providers.ftp.hooks.ftp import FTPSHook

        hook = FTPSHook("ftp_encoding")
        hook.get_conn()
        assert any(call.kwargs.get("encoding") == "cp1251" for call in mock_ftp_tls.mock_calls)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import socket
from unittest import mock

import pytest

from airflow.models import DAG, Connection
from airflow.providers.common.compat.openlineage.facet import Dataset
from airflow.providers.ftp.operators.ftp import (
    FTPFileTransmitOperator,
    FTPOperation,
    FTPSFileTransmitOperator,
)
from airflow.utils import timezone
from airflow.utils.timezone import datetime

DEFAULT_DATE = datetime(2017, 1, 1)
DEFAULT_CONN_ID = "ftp_default"


class TestFTPFileTransmitOperator:
    def setup_method(self):
        self.test_local_dir = "ftptmp"
        self.test_remote_dir = "/ftphome"
        self.test_remote_dir_int = "/ftphome/interdir"
        self.test_local_filename = "test_local_file"
        self.test_remote_filename = "test_remote_file"
        self.test_local_filepath = f"{self.test_local_dir}/{self.test_local_filename}"
        self.test_remote_filepath = f"{self.test_remote_dir}/{self.test_remote_filename}"
        self.test_local_filepath_int_dir = f"{self.test_local_dir}/{self.test_local_filename}"
        self.test_remote_filepath_int_dir = f"{self.test_remote_dir_int}/{self.test_remote_filename}"

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.store_file")
    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.create_directory")
    def test_file_transfer_put(self, mock_create_dir, mock_put):
        ftp_op = FTPFileTransmitOperator(
            task_id="test_ftp_put",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=self.test_local_filepath,
            remote_filepath=self.test_remote_filepath,
            operation=FTPOperation.PUT,
        )
        ftp_op.execute(None)
        assert not mock_create_dir.called
        mock_put.assert_called_once_with(self.test_remote_filepath, self.test_local_filepath)

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.store_file")
    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.create_directory")
    def test_file_transfer_with_intermediate_dir_put(self, mock_create_dir, mock_put):
        ftp_op = FTPFileTransmitOperator(
            task_id="test_ftp_put_imm_dirs",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=self.test_local_filepath,
            remote_filepath=self.test_remote_filepath_int_dir,
            operation=FTPOperation.PUT,
            create_intermediate_dirs=True,
        )
        ftp_op.execute(None)
        mock_create_dir.assert_called_with(self.test_remote_dir_int)
        mock_put.assert_called_once_with(self.test_remote_filepath_int_dir, self.test_local_filepath)

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.retrieve_file")
    def test_file_transfer_get(self, mock_get):
        ftp_op = FTPFileTransmitOperator(
            task_id="test_ftp_get",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=self.test_local_filepath,
            remote_filepath=self.test_remote_filepath,
            operation=FTPOperation.GET,
        )
        ftp_op.execute(None)
        mock_get.assert_called_once_with(self.test_remote_filepath, self.test_local_filepath)

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.retrieve_file")
    def test_file_transfer_with_intermediate_dir_get(self, mock_get, tmp_path):
        ftp_op = FTPFileTransmitOperator(
            task_id="test_ftp_get_imm_dirs",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=str(tmp_path / self.test_local_filepath_int_dir),
            remote_filepath=self.test_remote_filepath,
            operation=FTPOperation.GET,
            create_intermediate_dirs=True,
        )
        ftp_op.execute(None)
        assert len(list(tmp_path.iterdir())) == 1
        mock_get.assert_called_once_with(
            self.test_remote_filepath, str(tmp_path / self.test_local_filepath_int_dir)
        )

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.retrieve_file")
    def test_multiple_paths_get(self, mock_get):
        local_filepath = ["/tmp/ltest1", "/tmp/ltest2"]
        remote_filepath = ["/tmp/rtest1", "/tmp/rtest2"]
        ftp_op = FTPFileTransmitOperator(
            task_id="test_multiple_paths_get",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=local_filepath,
            remote_filepath=remote_filepath,
            operation=FTPOperation.GET,
        )
        ftp_op.execute(None)
        assert mock_get.call_count == 2
        for count, (args, _) in enumerate(mock_get.call_args_list):
            assert args == (remote_filepath[count], local_filepath[count])

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.store_file")
    def test_multiple_paths_put(self, mock_put):
        local_filepath = ["/tmp/ltest1", "/tmp/ltest2"]
        remote_filepath = ["/tmp/rtest1", "/tmp/rtest2"]
        ftp_op = FTPFileTransmitOperator(
            task_id="test_multiple_paths_put",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=local_filepath,
            remote_filepath=remote_filepath,
            operation=FTPOperation.PUT,
        )
        ftp_op.execute(None)
        assert mock_put.call_count == 2
        for count, (args, _) in enumerate(mock_put.call_args_list):
            assert args == (remote_filepath[count], local_filepath[count])

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.store_file")
    def test_arg_checking(self, mock_put):
        dag = DAG(
            dag_id="unit_tests_ftp_op_arg_checking",
            schedule=None,
            default_args={"start_date": DEFAULT_DATE},
        )
        # If ftp_conn_id is not passed in, it should be assigned the default connection id
        task_0 = FTPFileTransmitOperator(
            task_id="test_ftp_args_0",
            local_filepath=self.test_local_filepath,
            remote_filepath=self.test_remote_filepath,
            operation=FTPOperation.PUT,
            dag=dag,
        )
        task_0.execute(None)
        assert task_0.ftp_conn_id == DEFAULT_CONN_ID

        # Exception should be raised if operation is invalid
        task_1 = FTPFileTransmitOperator(
            task_id="test_ftp_args_1",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=self.test_local_filepath,
            remote_filepath=self.test_remote_filepath,
            operation="invalid_operation",
            dag=dag,
        )
        with pytest.raises(TypeError, match="Unsupported operation value invalid_operation, "):
            task_1.execute(None)

    def test_unequal_local_remote_file_paths(self):
        with pytest.raises(ValueError, match="1 paths in local_filepath != 2 paths in remote_filepath"):
            FTPFileTransmitOperator(
                task_id="test_ftp_unequal_paths",
                ftp_conn_id=DEFAULT_CONN_ID,
                local_filepath="/tmp/test",
                remote_filepath=["/tmp/test1", "/tmp/test2"],
            ).execute(None)

        with pytest.raises(ValueError, match="2 paths in local_filepath != 1 paths in remote_filepath"):
            FTPFileTransmitOperator(
                task_id="test_ftp_unequal_paths",
                ftp_conn_id=DEFAULT_CONN_ID,
                local_filepath=["/tmp/test1", "/tmp/test2"],
                remote_filepath="/tmp/test1",
            ).execute(None)


class TestFTPSFileTransmitOperator:
    def setup_method(self):
        self.test_local_dir = "ftpstmp"
        self.test_remote_dir = "/ftpshome"
        self.test_remote_dir_int = "/ftpshome/interdir"
        self.test_local_filename = "test_local_file"
        self.test_remote_filename = "test_remote_file"
        self.test_local_filepath = f"{self.test_local_dir}/{self.test_local_filename}"
        self.test_remote_filepath = f"{self.test_remote_dir}/{self.test_remote_filename}"
        self.test_local_filepath_int_dir = f"{self.test_local_dir}/{self.test_local_filename}"
        self.test_remote_filepath_int_dir = f"{self.test_remote_dir_int}/{self.test_remote_filename}"

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPSHook.store_file")
    @mock.patch("airflow.providers.ftp.operators.ftp.FTPSHook.create_directory")
    def test_file_transfer_put(self, mock_create_dir, mock_put):
        ftps_op = FTPSFileTransmitOperator(
            task_id="test_ftps_put",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=self.test_local_filepath,
            remote_filepath=self.test_remote_filepath,
            operation=FTPOperation.PUT,
        )
        ftps_op.execute(None)
        assert not mock_create_dir.called
        mock_put.assert_called_once_with(self.test_remote_filepath, self.test_local_filepath)

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPSHook.store_file")
    @mock.patch("airflow.providers.ftp.operators.ftp.FTPSHook.create_directory")
    def test_file_transfer_with_intermediate_dir_put(self, mock_create_dir, mock_put):
        ftps_op = FTPSFileTransmitOperator(
            task_id="test_ftps_put_imm_dirs",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=self.test_local_filepath,
            remote_filepath=self.test_remote_filepath_int_dir,
            operation=FTPOperation.PUT,
            create_intermediate_dirs=True,
        )
        ftps_op.execute(None)
        mock_create_dir.assert_called_with(self.test_remote_dir_int)
        mock_put.assert_called_once_with(self.test_remote_filepath_int_dir, self.test_local_filepath)

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPSHook.retrieve_file")
    def test_file_transfer_get(self, mock_get):
        ftps_op = FTPSFileTransmitOperator(
            task_id="test_ftps_get",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=self.test_local_filepath,
            remote_filepath=self.test_remote_filepath,
            operation=FTPOperation.GET,
        )
        ftps_op.execute(None)
        mock_get.assert_called_once_with(self.test_remote_filepath, self.test_local_filepath)

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPHook.retrieve_file")
    def test_file_transfer_with_intermediate_dir_get(self, mock_get, tmp_path):
        ftp_op = FTPFileTransmitOperator(
            task_id="test_ftp_get_imm_dirs",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=str(tmp_path / self.test_local_filepath_int_dir),
            remote_filepath=self.test_remote_filepath,
            operation=FTPOperation.GET,
            create_intermediate_dirs=True,
        )
        ftp_op.execute(None)
        assert len(list(tmp_path.iterdir())) == 1
        mock_get.assert_called_once_with(
            self.test_remote_filepath, str(tmp_path / self.test_local_filepath_int_dir)
        )

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPSHook.retrieve_file")
    def test_multiple_paths_get(self, mock_get):
        local_filepath = ["/tmp/ltest1", "/tmp/ltest2"]
        remote_filepath = ["/tmp/rtest1", "/tmp/rtest2"]
        ftps_op = FTPSFileTransmitOperator(
            task_id="test_multiple_paths_get",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=local_filepath,
            remote_filepath=remote_filepath,
            operation=FTPOperation.GET,
        )
        ftps_op.execute(None)
        assert mock_get.call_count == 2
        for count, (args, _) in enumerate(mock_get.call_args_list):
            assert args == (remote_filepath[count], local_filepath[count])

    @mock.patch("airflow.providers.ftp.operators.ftp.FTPSHook.store_file")
    def test_multiple_paths_put(self, mock_put):
        local_filepath = ["/tmp/ltest1", "/tmp/ltest2"]
        remote_filepath = ["/tmp/rtest1", "/tmp/rtest2"]
        ftps_op = FTPSFileTransmitOperator(
            task_id="test_multiple_paths_put",
            ftp_conn_id=DEFAULT_CONN_ID,
            local_filepath=local_filepath,
            remote_filepath=remote_filepath,
            operation=FTPOperation.PUT,
        )
        ftps_op.execute(None)
        assert mock_put.call_count == 2
        for count, (args, _) in enumerate(mock_put.call_args_list):
            assert args == (remote_filepath[count], local_filepath[count])

    @mock.patch("airflow.providers.ftp.hooks.ftp.FTPHook.get_conn", spec=Connection)
    def test_extract_get(self, get_conn):
        get_conn.return_value = Connection(
            conn_id="ftp_conn_id",
            conn_type="ftp",
            host="remotehost",
            port=21,
        )

        dag_id = "ftp_dag"
        task_id = "ftp_task"

        task = FTPFileTransmitOperator(
            task_id=task_id,
            ftp_conn_id="ftp_conn_id",
            dag=DAG(dag_id, schedule=None),
            start_date=timezone.utcnow(),
            local_filepath="/path/to/local",
            remote_filepath="/path/to/remote",
            operation=FTPOperation.GET,
        )
        lineage = task.get_openlineage_facets_on_start()

        assert lineage.inputs == [Dataset(namespace="file://remotehost:21", name="/path/to/remote")]
        assert lineage.outputs == [
            Dataset(
                namespace=f"file://{socket.gethostbyname(socket.gethostname())}:21", name="/path/to/local"
            )
        ]

    @mock.patch("airflow.providers.ftp.hooks.ftp.FTPHook.get_conn", spec=Connection)
    def test_extract_put(self, get_conn):
        get_conn.return_value = Connection(
            conn_id="ftp_conn_id",
            conn_type="ftp",
            host="remotehost",
            port=21,
        )

        dag_id = "ftp_dag"
        task_id = "ftp_task"

        task = FTPFileTransmitOperator(
            task_id=task_id,
            ftp_conn_id="ftp_conn_id",
            dag=DAG(dag_id, schedule=None),
            start_date=timezone.utcnow(),
            local_filepath="/path/to/local",
            remote_filepath="/path/to/remote",
            operation=FTPOperation.PUT,
        )
        lineage = task.get_openlineage_facets_on_start()

        assert lineage.inputs == [
            Dataset(
                namespace=f"file://{socket.gethostbyname(socket.gethostname())}:21", name="/path/to/local"
            )
        ]
        assert lineage.outputs == [Dataset(namespace="file://remotehost:21", name="/path/to/remote")]
